/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef FCGID_BRIDGE_H
#define FCGID_BRIDGE_H
#include "httpd.h"
#include "ap_config.h"
#include "http_config.h"
#include "apr_hash.h"
#include "fcgid_conf.h"

apr_status_t bucket_ctx_cleanup(void *thectx);
int bridge_request(request_rec * r, int role, fcgid_cmd_conf *cmd_conf);

#endif
